#! /usr/bin/python
# -*- coding: utf-8 -*-
################################################################################
# This script converts *.spe files produced by CCD camera into a ROOT file.    #
# Each frame is saved as an instance of TGraph class                           #
#                                                                              #
# requirements: python 2.6 with library NumPy                                  #
# usage: python spe2root file.spe                                              #
# version 0.1 (2010/08/06)                                                     #
#                                                                              #
# This script is based on work of Andre Espaze,                                #
# see http://www.scipy.org/Cookbook/Reading_SPE_files for details              #
#                                                                              #
# Created by Vladimir Jary <Vladimir.Jary@cern.ch> for OSQAR collaboration     #
#                                                                              #
################################################################################

import numpy as N
from array import array
from ROOT import gROOT, gDirectory, gPad, gSystem, gRandom 
from ROOT import TObject,TGraph,TFile
import spe2root


# class representing SPE file
class File(object):

    #opens file
    def __init__(self, fname):
        self._fid = open(fname, 'rb')
        self._load_header()

    #reads information about SPE file from its header
    def _load_header(self):
      self._xdim = N.int64(self.read_at(42, 1, N.int16)[0])
      self._frames = N.int64(self.read_at(34, 1, N.int16)[0])
      self._datatype = N.int64(self.read_at(108, 1, N.int16)[0])
      self._ydim = N.int64(self.read_at(656, 1, N.int16)[0])
  
    #def get_size(self):
    #    return (self._xdim, self._ydim)
    
    #reads data of type ntype from SPE file at position pos
    def read_at(self, pos, size, ntype):
        self._fid.seek(pos)
        return N.fromfile(self._fid, ntype, size)

    #reads frames from SPE file
    def load_img(self):
      if self._datatype == 0:
        img = self.read_at(4100, self._xdim * self._ydim * self._frames, N.float32)
      elif self._datatype == 1:
        img = self.read_at(4100, self._xdim * self._ydim * self._frames, N.int32)
      elif self._datatype == 2:
        img = self.read_at(4100, self._xdim * self._ydim * self._frames, N.int16)
      elif self._datatype == 3:
	img = self.read_at(4100, self._xdim * self._ydim * self._frames, N.uint16)
        #return img.reshape((self._ydim, self._xdim))
      return img
    
    #closes SPE file
    def close(self):
        self._fid.close()

#generates name of output file
def construct_name(input):
	name = input.lower()
	name = name.replace(".spe", ".root")
	return name

#Writes frame as a TGraph containing data xdata, ydata into root file
def write_graph(xdata, ydata, name, frame):
    #create graph
    graph = TGraph(len(xdata), xdata, ydata);
    #prepare title of graph
    title = "";
    if frame < 10:
      title = "000"
    elif frame < 100:
      title = "00"
    elif frame < 1000:
      title = "0"
    title = title + str(frame)
    #set title and name  of graph
    graph.SetTitle("File: " + name + ", Frame " + title)
    graph.SetName("Frame_" + title)
    #write graph into root file
    graph.Write()
    print "Written frame " + title
    

#converts spe file fname into ROOT file
def convert(fname):
    #open input SPE file
    fid = File(fname)
    #open output ROOT file
    root = TFile(construct_name(fname), "recreate")
    #load frames
    img = fid.load_img()
    #parse header information
    columns = fid._xdim
    rows = fid._ydim
    frames = fid._frames
    #number of pixels in frame
    pixels = rows*columns
    #print header
    print "Input file: " + fname
    print "Number of columns: " + str(columns)
    print "Number of rows: " + str(rows)
    print "Number of pixels in frame: " + str(pixels)
    print "Number of frames: " + str(frames)
    print "Data type: " + str(fid._datatype)
    
    #array of pixel numbers
    xx = array( 'f', [] )
    #array of pixel values
    yy = array( 'f', [] )
    #initialize arrays
    for i in range(0, pixels):
      xx.append(i+1)
      yy.append(0)
    
    #read pixel intensities pixel by pixel
    for frame in range(0, frames):
      for i in range(0, pixels):
	yy[i]=img[i+pixels*frame]
      write_graph(xx, yy, fname, frame+1)
    #close files
    root.Write()
    root.Close()
    fid.close()
    
if __name__ == "__main__":
    import sys
    convert(sys.argv[-1])
