/***************************************************************************
 *   Copyright (C) 2010 by Vladimir Jary                                   *
 *   <Vladimir.Jary@cern.ch>                                               *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 ***************************************************************************/

#include "speexporter.h"

#include <QtCore/QFile>
#include <QtCore/QString>
#include <QtCore/QIODevice>
#include <QtCore/QDataStream>
#include <QtCore/QTextStream>
#include <QtCore/QStringList>

#ifdef USE_ROOT
#include <TROOT.h>
#include <TFile.h>
#include <TGraph.h>
#include <TF1.h>
#include <TCanvas.h>
#include <TAxis.h>
#endif

#include "iostream"


/*
  Initialization of input and output streams and parses
  header of input SPE file
*/
SpeExporter::SpeExporter(QString input)
{
    inFile = new QFile();
    outFile = new QFile();
    inputName = input;
    inFile->setFileName(input);
    inFile->open(QIODevice::ReadOnly);
    in = new QDataStream(inFile);
    in->setFloatingPointPrecision(QDataStream::SinglePrecision);
    out = new QTextStream;
    newLine = "\n";
    setNewLine(UNIX);
    parseHeader();
    separator=';';
}

/*
  Closes and deletes specified file
*/
void SpeExporter::deleteFile(QFile *file)
{
    if (file){
        if (file->isOpen()) file->close();
        delete file;
        file = 0;
    }
}

/*
   Clean-up, destroys files
 */
SpeExporter::~SpeExporter()
{
    deleteFile(inFile);
    deleteFile(outFile);
}

/*
  Sets the newline sequence according to desired platform:
  DOS/Windows: CR, LF
  Unix: LF
  Macintosh: CR
  see http://en.wikipedia.org/wiki/Newline
*/
void SpeExporter::setNewLine(Platform platform)
{
    if(platform == DOS)
        newLine = "\r\n"; //DOS
    else if(platform == MAC)
        newLine = "\r"; //Macintosh
    else newLine = "\n"; //Unix
}

/*
  Prints relevant information from SPE file header:
  frame dimension, number of frames, and data format
*/
void SpeExporter::printHeader()
{
    std::cout << "-----------------------" << std::endl;
    std::cout << "Input: " << qPrintable(inputName) << std::endl;
    std::cout << "x: " << x << std::endl;
    std::cout << "y: " << y << std::endl;
    std::cout << "Size: " << size << " " << std::endl;
    std::cout << "Number of frames: " << frames << std::endl;
    std::cout << "Data format: " << format << " (";
    switch (format){
    case FLOAT: std::cout << "32bit float";
            break;
    case INT32: std::cout << "32bit integer";
            break;
    case INT16: std::cout << "16bit signed integer - currently untested!";
        break;
    case UINT16: std::cout << "16bit unsigned integer";
        break;
    default:
        std::cout << "unknown format";
        break;
    }
    std::cout << ")" << std::endl;
    std::cout << "-----------------------" << std::endl;
}

/*
  Processes header of opened SPE file
  x ... stored at position 42
  y ... stored at position 656
*/
void SpeExporter::parseHeader()
{

    format = UINT16;
    in->skipRawData(34);
    //frames = readUInt16(in);
    readPixel(in, frames);

    //if(frames == 65535) frames = 1;

    in->skipRawData(6);
    //x = readUInt16(in);
    readPixel(in, x);

    in->skipRawData(64);
    //format = readUInt16(in);
    readPixel(in, format);

    in->skipRawData(546);
    //y = readUInt16(in);
    readPixel(in, y);
    size = x*y;

    printHeader();

    //skip rest of header
    in->skipRawData(3442);

}

/*
  Reads one pixel intensity from input stream instream
*/
template<typename TPixelFormat> void SpeExporter::readPixel(QDataStream* instream, TPixelFormat &intensity)
{
    if(format == UINT16){
        qint8 byte1, byte2;
        qint32 b1, b2;
        instream->setByteOrder(QDataStream::LittleEndian);
        instream->operator >>(byte1);
        instream->operator >>(byte2);

        b1 = byte1;
        b2 = byte2;
        if (b1 < 0) b1+=256;
        if (b2 < 0) b2+=256;
        intensity = (b2 << 8) | b1;
    } else
    instream->operator >> (intensity);
}


/*
  Prepares input and output streams
 */
void SpeExporter::openStreams()
{
    in->setDevice(inFile);
    outFile->setFileName(outputName);
    outFile->open(QIODevice::WriteOnly);
    out->setDevice(outFile);
}

/*
  Closes opened files
 */
void SpeExporter::closeStreams()
{
    inFile->close();
    outFile->close();
}

/*
  Converts input SPE file into format specified by the extension
*/
bool SpeExporter::exportFile(QString ext)
{
    float f = 0.0;
    qint32 i32 = 0;
    qint16 i16 = 0;
    quint16 ui16 = 0;
    //plain text
    if(ext == "txt"){
        switch(format){
        case FLOAT:
            export2txt(f);
            break;
        case INT32:
            export2txt(i32);
            break;
        case INT16:
            export2txt(i16);
            break;
        case UINT16:
            export2txt(ui16);
            break;
        default:
            std::cout << "Unknown data format" << std::endl;
            return false;
        };
    }
    //comma-separated values
    else if(ext == "csv")
    {
        switch(format){
        case FLOAT:
            export2csv(f);
            break;
        case INT32:
            export2csv(i32);
            break;
        case INT16:
            export2csv(i16);
            break;
        case UINT16:
            export2csv(ui16);
            break;
        default:
            std::cout << "Unknown data format" << std::endl;
            return false;
        };
    }
    //ASCII files
    else if(ext == "asc")
    {
        switch(format){
        case FLOAT:
            export2asc(f);
            break;
        case INT32:
            export2asc(i32);
            break;
        case INT16:
            export2asc(i16);
            break;
        case UINT16:
            export2asc(ui16);
            break;
        default:
            std::cout << "Unknown data format" << std::endl;
            return false;
        };
    }
    //ROOT file
#ifdef USE_ROOT
    else if(ext == "root")
    {
        switch(format){
        case FLOAT:
            export2root(f);
            break;
        case INT32:
            export2root(i32);
            break;
        case INT16:
            export2root(i16);
            break;
        case UINT16:
            export2root(ui16);
            break;
        default:
            std::cout << "Unknown data format" << std::endl;
            return false;
        };

    }
#endif
    else return false;
    return true;
}

/*
  Allocates dynamic 2D array with given rows and columns
*/
//void SpeExporter::Allocate2DArray(qint32** &buff, quint32 rows, quint32 cols)
template<typename TPixelFormat> void SpeExporter::Allocate2DArray(TPixelFormat** &buff, quint32 rows, quint32 cols)
{
    //buff = new qint32*[rows];
    buff = new TPixelFormat*[rows];
    for(quint32 i = 0; i < rows; i++)
    {
        //buff[i] = new qint32[cols];
        buff[i] = new TPixelFormat[cols];
    }
}

/*
  Releases memory occupied by 2D array buff
*/
//void SpeExporter::Free2DArray(qint32** &buff)
template<typename TPixelFormat> void SpeExporter::Free2DArray(TPixelFormat** &buff)
{
    delete [] *buff;
    delete [] buff;
    buff = 0;
}


/*
  Loads data from input stream in into 2D array buff
*/
//void SpeExporter::loadData(qint32**buff, QDataStream *in)
template<typename TPixelFormat> void SpeExporter::loadData(TPixelFormat** buff, QDataStream *in)
{

    //qint32 intensity;
    TPixelFormat intensity;
    for(quint16 i = 0; i < frames; i++)
    {
        for(quint32 j = 0; j < size; j++){
            readPixel(in, intensity);
            buff[i][j] = intensity;
         }
    }

}

/*
  Converts SPE file into plain-text format
 */
//void SpeExporter::export2txt()
template<typename TPixelFormat> void SpeExporter::export2txt(TPixelFormat intensity)
{
    intensity = 0;
    constructOutputName(".txt");
    openStreams();

    TPixelFormat **buffer = 0;
    Allocate2DArray(buffer, frames, size);
    loadData(buffer, in);

    for(quint32 i = 0; i < size; i++)
    {
        for(quint32 j = 0; j < frames; j++){
            out->operator << (buffer[j][i]);
            if(j+1 == frames) out->operator <<(newLine);
                else out->operator <<(' ');
        }
    }

    Free2DArray(buffer);
    closeStreams();
}


/*
  Converts input SPE file into comma-separated value
  spreadsheet
 */
//void SpeExporter::export2csv()
template <typename TPixelFormat> void SpeExporter::export2csv(TPixelFormat intensity)
{
    intensity = 0;
    constructOutputName(".csv");
    openStreams();

    TPixelFormat **buffer = 0;
    Allocate2DArray(buffer, frames, size);
    loadData(buffer, in);

    for(quint32 i = 0; i < size; i++)
    {
        //print pixel number
        out->operator <<(i+1) << separator;
        for(quint32 j = 0; j < frames; j++){
            out->operator << (buffer[j][i]);
            if(j+1 == frames) out->operator <<(newLine);
                else out->operator <<(separator);
        }
    }

    Free2DArray(buffer);
    closeStreams();
}


/*
  Converts SPE file into plain-text format, each scan (column)
  is stored into separate ASCII file (*.asc)
*/
//void SpeExporter::export2asc()
template <typename TPixelFormat> void SpeExporter::export2asc(TPixelFormat intensity)
{
    quint16 scan = 0;
    QString suffix;


    for(quint32 j = 0; j < size*frames; j++){
        if(j % size == 0)
        {
            outFile->close();
            scan+=1;
            suffix.setNum(scan);
            if(scan < 10)
                suffix.prepend("_000");
            else if(scan < 100)
                suffix.prepend("_00");
            else if(scan < 1000)
                suffix.prepend("_0");
            else suffix.prepend("_");

            suffix.append(".asc");
            constructOutputName(suffix);


            //std::cout << scan <<  " " << qPrintable(inputName) << " " << qPrintable(outputName) << std::endl;
            openStreams();
        }

        readPixel(in, intensity);
        out->operator << (intensity) << (newLine);

    }
    closeStreams();
}


#ifdef USE_ROOT
/*
  Converts SPE file into root file.
  Each frame is saved as an instance of the TGraph class
*/
template <typename TPixelFormat> void SpeExporter::export2root(TPixelFormat intensity)
//void SpeExporter::export2root()
{
    intensity = 0;
    constructOutputName(".root");
    //create root file
    TFile *root = new TFile(qPrintable(outputName), "recreate");
    QString title;
    QString number;
    TPixelFormat **buffer = 0;
    Allocate2DArray(buffer, frames, size);
    loadData(buffer, in);

    //x axis data
    float x[size];
    for(uint i = 0; i < size; i++)
           x[i] = i + 1;
    //y values
    float y[size];


    for(int i = 0; i < frames; i++)
    {
        for(uint j = 0; j < size; j++)
           y[j]=(float)buffer[i][j];
        //create graph with given data points
        TGraph *graph = new TGraph(size, x, y);

        //line colour
        graph->SetLineColor(2);
        //thickness of line
        graph->SetLineWidth(1);


        if(i < 9)
            number = "000";
        else if(i < 99)
            number = "00";
        else if(i < 999)
            number = "0";
        else number = "";

        number = number + QString("%1").arg(i + 1);
        //graph title
        title=QString("File %1, frame %2").arg(inputName).arg(number);
        graph->SetTitle(qPrintable(title));
        //graph name
        title=QString("Frame_%1").arg(number);
        graph->SetName(qPrintable(title));
        //axis
        graph->GetXaxis()->SetTitle("Pixel number");
        graph->GetYaxis()->SetTitle("Intensity");
        //write graph into the root file
        graph->Write();
        //delete graph from memory
        delete graph;
   }

   //clean nup
   Free2DArray(buffer);
   root->Close();
}
#endif

/*
  Generates name for output file - same as input, but in lowercase
  and with changed extension (SPE -> csv/txt/root)
 */
void SpeExporter::constructOutputName(QString extension)
{
  //outputName = inputName.toLower();
    QString tmp = inputName;
    outputName = inputName.replace(".spe", extension, Qt::CaseInsensitive);
    inputName = tmp;
}
